# -*- coding: utf-8 -*-
# (c) 2017,2018 Andreas Motl <andreas@hiveeyes.org>
# (c) 2017,2018 Richard Pobering <richard@hiveeyes.org>
# License: GNU General Public License, Version 3

# This program has moved to https://github.com/hiveeyes/luftdatenpumpe
# Please update your links and bookmarks.

if __name__ == '__main__':
    print('This program has moved to https://github.com/hiveeyes/luftdatenpumpe')
